#include "usage_wrapper.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
/* SIZE: buffer size = SIZE*work_index */
#define SIZE      1024*1024
/* NUM_WORKS: number of do_works done */
#define NUM_WORKS 5
/* some time spend doing something in do_work */
#define LOOPS     30
/* -----------------------------------------------------------------
 * program that loops through "work" routines (allocates space) then
 * frees the memory.
 * -----------------------------------------------------------------
 */
void do_work(
             int buf_size_factor,
             char **buf
             );
int main(
         int argc,
         char **argv
         )
{
  int i;                 /* loop variable */
  int ierr;              /* error return value */
  int pinfo_print;       /* whether or not to print stuff */
  double p_wall[2];      /* secs wall clock time (cur, delta) */
  double p_time[2];      /* secs proc time (cur, delta) */
  double p_size[2];      /* byte proc size (cur, delta) */
  double size_tot;       /* total size allocated */
  char *bufs[NUM_WORKS]; /* array of buffs allocated by do_work */
  /* print every time USAGE_WRAPPER is called */
  pinfo_print = 1;
  /* print header */
  printf( "\n%10s [%10s] %10s [%10s] %10s [%10s] : [%10s]\n",
          "p_wall now", "delta",
          "p_time now", "delta",
          "p_size now", "delta",
          "expected");
  /* loop of get info , do_work, get info, print changes */
  size_tot = 0;
  for( i = 1; i <= NUM_WORKS; i++ )
    {
      size_tot += i*SIZE;
      USAGE_WRAPPER( &pinfo_print, p_wall, p_time, p_size, &ierr );
      do_work( i, &(bufs[i-1]) );
      USAGE_WRAPPER( &pinfo_print, p_wall, p_time, p_size, &ierr );
      printf( "%10.2e [%10.2e] %10.2e [%10.2e] %10.2e [%10.2e] : [%10.2e]\n",
              p_wall[0], p_wall[1],
              p_time[0], p_time[1],
              p_size[0], p_size[1],
              i*SIZE*1.0 );
    }
  printf( "Free Memory\n" );
  /* free buffers */
  USAGE_WRAPPER( &pinfo_print, p_wall, p_time, p_size, &ierr );
  for( i = 0; i < NUM_WORKS; i++ )
    {
      free( bufs[i] );
      bufs[i] = NULL;
    }
  USAGE_WRAPPER( &pinfo_print, p_wall, p_time, p_size, &ierr );
  printf( "%10.2e [%10.2e] %10.2e [%10.2e] %10.2e [%10.2e] : [%10.2e]\n",
          p_wall[0], p_wall[1],
          p_time[0], p_time[1],
          p_size[0], p_size[1],
          size_tot*(-1.0) );
  printf( "\n" );
  return( ierr );
}
/* allocate space and spend time doing something */
void do_work(
             int buf_size_factor,
             char **buf
             )
{
  int i;            /* loop var */
  int j;            /* loop var */
  /* allocate space */
  *buf = calloc( SIZE*buf_size_factor, sizeof( char ) );
  /* spend time doing something */
  for( j = 0; j < LOOPS; j++ )
    {
      for( i = 0; i < SIZE*buf_size_factor; i++ )
        {
          (*buf)[i] = 'a';
        }
    }
}
