#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <signal.h>
#include <unistd.h>
#include <assert.h>
/* ------------- */
/* include files */
/* ------------- */
#if defined( _PROCMON_UPS )
#include "ups.h"
#else
#if defined( _PROCMON_SERIAL )
#else
#include "mpi.h"
#endif
#endif
/* ITERATIONS = number of outer loops */
#define ITERATIONS 500
/* NUM_SAVES = number of buffers obtained before freeing */
#define NUM_SAVES 10000
/* ------------------------------------------------------------------ */
/* each process loops through creating buffers and pe 0 does not free */
/* ------------------------------------------------------------------ */
int main(int argc, char **argv)
{
  int
    i, /* loop var */
    j, /* loop var */
    ierr = 0, /* error return value */
    penum=0; /* process number */
  char
    *buf[NUM_SAVES]; /* leaky char array */

  /* init and get process number */
#if defined( _PROCMON_UPS )
  ierr = UPS_AA_Init( argc, argv );
  assert( !ierr );
  ierr = UPS_CM_Get_penum( &penum );
  assert( !ierr );
#else
#if defined( _PROCMON_SERIAL )
  penum = 0;
#else
  ierr = MPI_Init( &argc, &argv );
  assert( !ierr );
  ierr = MPI_Comm_rank( MPI_COMM_WORLD, &penum );
  assert( !ierr );
#endif
#endif
  fprintf( stderr, "parallel_leak: %d says hi\n", penum );
  /* --------------------------------------------------------- */
  /* loop doing allocation and some busy work to trigger timer */
  /* and create mem leak.                                      */
  /* --------------------------------------------------------- */
  for( i = 0; i < ITERATIONS; i++ )
    {
      for( j = 0; j < NUM_SAVES; j++ )
        {
          buf[j] = malloc( getpagesize() );
          if( buf == NULL )
            {
              fprintf( stderr, "penum [%d] malloc failed\n", penum );
              ierr = 1;
              break;
            }
          memset( buf[j], 1, getpagesize() );
          memset( buf[j], 2, getpagesize() );
        }
      if( penum != 0 )
        {
          for( j = 0; j < NUM_SAVES; j++ )
            {
              free( buf[j] );
            }
        }
      else
        {
          fprintf( stderr, "leak %d/%d\n", i+1, ITERATIONS );
        }
#if defined( _PROCMON_SYNC )
#if defined( _PROCMON_UPS )
      ierr = UPS_CM_Barrier();
      assert( !ierr );
#else
#if defined( _PROCMON_SERIAL )
#else
      ierr = MPI_Barrier( MPI_COMM_WORLD );
      assert( !ierr );
#endif
#endif
#endif
    }
  /* terminate */
#if defined( _PROCMON_UPS )
  ierr = UPS_AA_Terminate();
  assert( !ierr );
#else
#if defined( _PROCMON_SERIAL )
  penum = 0;
#else
  ierr = MPI_Finalize();
  assert( !ierr );
#endif
#endif
  return( ierr );
}
