#include "procmon_info.h"
#include <unistd.h>
#include <time.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <stdlib.h>
#include <ctype.h>
typedef struct PROCMON_INFO_LIST_struct
{
  struct PROCMON_INFO_LIST_struct *next;
  PROCMON_INFO_struct pinfo;
  FILE *outfile; /* output file pointer */
  char outfile_name[PROCMON_MAX_LEN]; /* output file name */
  long long pid; /* process id for this struct */
}
PROCMON_INFO_LIST_struct;
int procmon_sleep(
                  double time
                  );
int procmon_list_create(
                        PROCMON_INFO_SETTINGS_struct *psettings,
                        PROCMON_INFO_LIST_struct **plist_head )
{
  int ierr=0; /* error return variable */
  int match; /* T/F it search process matches */
  DIR *proc_dir; /* opendir */
  struct dirent *proc_dir_dirent; /* readdir */
  char cwd[PROCMON_MAX_LEN]; /* current working dir */
  long long pid; /* process id for search in /proc */
  PROCMON_INFO_struct pinfo; /* temp for search in /proc */
  PROCMON_INFO_LIST_struct *plist; /* pointer to list */
  /* ----------------------------------------- */
  /* if getting info about a list, create list */
  /* ----------------------------------------- */
  if( psettings->pid == 0 )
    {
      getcwd( cwd, PROCMON_MAX_LEN );
      proc_dir = opendir("/proc");
      while( (proc_dir_dirent = readdir(proc_dir)) != NULL )
        {
          if( isdigit( proc_dir_dirent->d_name[0] ) )
            {
              pid = atoi( proc_dir_dirent->d_name );
              procmon_info_init( &pinfo );
              procmon_info_get( pid, &pinfo );
              match = 0;
              if( strlen( psettings->command ) == 0 &&
                  psettings->uid >= 0 &&
                  pinfo.p_uid == psettings->uid )
                {
                  match = 1;
                }
              if( psettings->uid < 0 &&
                  strlen( psettings->command ) > 0 &&
                  strstr( pinfo.p_command, psettings->command ) != NULL )
                {
                  match = 1;
                }
              if( psettings->uid >= 0 &&
                  pinfo.p_uid == psettings->uid &&
                  strlen( psettings->command ) > 0 &&
                  strstr( pinfo.p_command, psettings->command ) != NULL )
                {
                  match = 1;
                }
              if( match != 0 )
                {
                  plist = calloc( 1, sizeof( PROCMON_INFO_LIST_struct ) );
                  plist->pid = pid;
                  if( strlen( psettings->outfile_prefix ) > 0 )
                    {
                      sprintf( plist->outfile_name, "%s%llu%s",
                               psettings->outfile_prefix, plist->pid, ".txt" );
                      plist->outfile = fopen( plist->outfile_name, "a" );
                    }
                  else
                    {
                      plist->outfile = NULL;
                    }
                  plist->next = *plist_head;
                  *plist_head = plist;
                }
            }
        }
    }
  /* ----------------------------------------------- */
  /* DONE: if getting info about a list, create list */
  /* ----------------------------------------------- */
  /* ------------------------------------------------------ */
  /* if getting info about a single pe, just create 1 entry */
  /* ------------------------------------------------------ */
  else
    {
      plist = calloc( 1, sizeof( PROCMON_INFO_LIST_struct ) );
      plist->pid = psettings->pid;
      procmon_info_init( &(plist->pinfo) );
      if( strlen( psettings->outfile_prefix ) > 0 )
        {
          sprintf( plist->outfile_name, "%s%llu%s",
                   psettings->outfile_prefix, plist->pid, ".txt" );
          plist->outfile = fopen( plist->outfile_name, "a" );
        }
      else
        {
          plist->outfile = NULL;
        }
      *plist_head = plist;
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   procmon

   Purpose
   =======
   Standalone tool used somewhat like ps to get process info.

   Arguments
   =========
   argc          Intent: in
                 C type: int
                 Argument count.

   argv          Intent: in
                 C type: char**
                 Arguments

   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) get settings
   2) set outfile
   3) print settings and header
   4) gather/print samples
   4.1) procmon_info_get
   4.2) if- procmon_info_print
   4.3) procmon_sleep

   ======================================================================== */
int main( int argc, const char **argv )
{
  /* --------- */
  /* variables */
  /* --------- */
  int ierr = 0; /* error return value */
  int ierr_pget; /* error for procmon_info_get */
  PROCMON_INFO_SETTINGS_struct psettings; /* settings */
  PROCMON_INFO_LIST_struct *plist_head; /* points to head of list */
  PROCMON_INFO_LIST_struct *plist_previous; /* points to previous in list */
  PROCMON_INFO_LIST_struct *plist_next; /* points to next in list */
  PROCMON_INFO_LIST_struct *plist; /* points to a member of list */
  /* --------------------- */
  /* executable statements */
  /* --------------------- */
  if( argc == 1 )
    {
      printf( "\n" );
      printf(
             "Usage: procmon <pid> [options]\n"
             "\n"
             "Options:\n"
             "   -time <millisecs>\n"
             "      Time interval between samplings.\n"
             "      Default: (0) Sample and print once instead of sampling\n"
             "               until process ends.\n"
             "   -dir <output dir>\n"
             "      Outpt directory for files.\n"
             "      File Format: <output dir>/procmon_<machine>_<pid>\n"
             "      Default: Print to stdout.\n"
             "   -interval <n>\n"
             "      Print every <n-th> sampling.\n"
             "      Default: (1) Print every sampling.\n"
             "      If set to 0, only print at process termination.\n"
             "   -uid <n>\n"
             "      Gather info about all processes with the uid.\n"
             "      Processes that start after this command is executed\n"
             "      will not be sampled.\n"
             "      Options -uid and -command will be intersected.\n"
             "   -command <command>\n"
             "      Gather info about all processes with the command.\n"
             "      Processes that start after this command is executed\n"
             "      will not be sampled.\n"
             "      Options -uid and -command will be intersected.\n"
             "Post Processing:\n"
             "   When finished, procmon_plot.pl can be used to generate\n"
             "   plots of the output file(s).  For example:\n"
             "      procmon_plot.pl procmon_*.txt\n"
             "\n"
             );
      return( 0 );
    }
  /* ------------ */
  /* get settings */
  /* ------------ */
  ierr = procmon_info_settings_get( argc, argv, &psettings );
  assert( !ierr );
  /* ----------------- */
  /* create pinfo list */
  /* ----------------- */
  plist_head = NULL;
  procmon_list_create( &psettings, &plist_head );
  /* -------------- */
  /* print settings */
  /* -------------- */
  plist = plist_head;
  while( plist != NULL )
    {
      procmon_info_settings_print( &psettings, plist->outfile );
      plist = plist->next;
    }
  /* -------------------- */
  /* gather/print samples */
  /* -------------------- */
  while( plist_head != NULL )
    {
      /* ------------ */
      /* process list */
      /* ------------ */
      plist = plist_head;
      plist_previous = NULL;
      while( plist )
        {
          ierr_pget = procmon_info_get( plist->pid, &(plist->pinfo) );
          /* ------------------------------- */
          /* print if:                       */
          /* - interval and good data        */
          /* - time == 0 (only printing one) */
          /* - PROCMON_ERROR (process gone)  */
          /* ------------------------------- */
          if( ( ierr_pget == 0 &&
                psettings.interval > 0 &&
                plist->pinfo.num_samples % psettings.interval == 0 ) ||
              ( psettings.time == 0 ) ||
              ( ierr_pget == PROCMON_ERROR ) )
            {
              ierr = procmon_info_print( &(plist->pinfo), (plist->outfile) );
            }
          /* ------------------------ */
          /* if this is last sampling */
          /* ------------------------ */
          if( psettings.time == 0 || ierr_pget == PROCMON_ERROR )
            {
              /* close file */
              if( plist->outfile != NULL )
                {
                  fclose( plist->outfile );
                  plist->outfile = NULL;
                  if( plist->pinfo.num_samples < psettings.min )
                    {
                      ierr = remove( plist->outfile_name );
                      assert( !ierr );
                    }
                }
              /* free from list */
              if( plist == plist_head )
                {
                  plist_head = plist->next;
                }
              else
                {
                  plist_previous->next = plist->next;
                }
              plist_next = plist->next;
              free( plist );
              plist = plist_next;
              continue;
            }
          plist_previous = plist;
          plist = plist->next;
        }
      ierr = procmon_sleep( psettings.time/1000.0 );
    }
  /* -------------------------- */
  /* DONE: gather/print samples */
  /* -------------------------- */
  return( ierr );
}
/* ========================================================================

   Name
   ====
   procmon_sleep

   Purpose
   =======
   fine-grained sleep timer

   Arguments
   =========
   time          Intent: in
                 C type: double
                 time in seconds to sleep.

   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) sleep

   ======================================================================== */
int procmon_sleep( double time )
{
  int ierr = 0; /* error return value */
#if defined( _POSIX_TIMERS )
  struct timespec sleep_time_fine; /* sent to fine sleep timer */
#else
  useconds_t sleep_time_fine; /* sent to fine sleep timer */
#endif
  int sleep_time_secs; /* secs to sleep */
  int sleep_time_usecs; /* usecs to sleep */
  /* -------------------------------------- */
  /* creat secs and part and sleep for secs */
  /* -------------------------------------- */
  sleep_time_secs = (int)time;
  sleep_time_usecs = (time - sleep_time_secs) * 1000000.0;
  if( sleep_time_secs > 0 )
    {
      sleep( sleep_time_secs );
    }
  /* -------------- */
  /* sleep for part */
  /* -------------- */
#if defined( _POSIX_TIMERS )
  sleep_time_fine.tv_sec = 0;
  sleep_time_fine.tv_nsec = sleep_time_usecs * 1000;
  nanosleep( &sleep_time_fine, NULL );
#else
  usleep( sleep_time_usecs );
#endif
  return( ierr );
}
