#include "procmon_info.h"
#include <stdlib.h>
#include <fcntl.h>
#include <sys/procfs.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#define PROCMON_TIMEN_TO_DOUBLE(t) \
  ((double)((t).tv_sec) + (double)((t).tv_nsec) / 1000000000.0)
/* ========================================================================

   Name
   ====
   procmon_info_settings_get

   Purpose
   =======
   Get the settings for procmon.  This will process command line
   args or environment variable settings.

   Arguments
   =========
   argc       Intent: in
              C type: int
              Argument count (from main()).
              If 0, will just process environment vars.
              
   argv       Intent: in
              C type: const char**
              Argument list (from main()).
              If NULL, will just process environment vars.

   psettings  Intent: in
              C type: PROCMON_INFO_SETTINGS_struct*
              Struct to be filled.

   ierr       Intent: out
              C type: (na) error return value
              error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) get env var settings.
   2) override with any command line settings

   ======================================================================== */
int procmon_info_settings_get(
                              int argc,
                              const char **argv,
                              PROCMON_INFO_SETTINGS_struct *psettings
                              )
{
  int
    i, /* loop var */
    ierr=0; /* error return value */
  char
    *empty_string = "", /* empty string */
    *getenv_char_p, /* pointer to env var */
    hostname[PROCMON_MAX_LEN]; /* the hostname */
  /* ---- */
  /* init */
  /* ---- */
  /* -------- */
  /* defaults */
  /* -------- */
  psettings->time     = 0;
  psettings->interval = 1;
  psettings->min      = 0;
  sprintf( psettings->dir, "%s", empty_string );
  psettings->pid      = 0;
  psettings->uid      = -1;
  sprintf( psettings->command, "%s", empty_string );
  /* -------------------- */
  /* get env var settings */
  /* -------------------- */
  /* ------------ */
  /* PROCMON_TIME */
  /* ------------ */
  getenv_char_p = getenv( "PROCMON_TIME" );
  if( getenv_char_p != NULL )
    {
      psettings->time = atoi( getenv_char_p );
      if( psettings->time < 0 )
        {
          psettings->time = 0;
        }
    }
  /* ---------------- */
  /* PROCMON_INTERVAL */
  /* ---------------- */
  getenv_char_p = getenv( "PROCMON_INTERVAL" );
  if( getenv_char_p != NULL )
    {
      psettings->interval = atoi( getenv_char_p );
      if( psettings->interval < 0 )
        {
          psettings->interval = 0;
        }
    }
  /* ----------- */
  /* PROCMON_MIN */
  /* ----------- */
  getenv_char_p = getenv( "PROCMON_MIN" );
  if( getenv_char_p != NULL )
    {
      psettings->min = atoi( getenv_char_p );
      if( psettings->min < 0 )
        {
          psettings->min = 0;
        }
    }
  /* ----------- */
  /* PROCMON_DIR */
  /* ----------- */
  getenv_char_p = getenv( "PROCMON_DIR" );
  if( getenv_char_p == NULL || strlen( getenv_char_p ) <= 0 )
    {
      getenv_char_p = empty_string;
    }
  assert( strlen(getenv_char_p) < PROCMON_MAX_LEN );
  sprintf( psettings->dir, "%s", getenv_char_p );
  /* --------------- */
  /* PROCMON_COMMAND */
  /* --------------- */
  getenv_char_p = getenv( "PROCMON_COMMAND" );
  if( getenv_char_p == NULL || strlen( getenv_char_p ) <= 0 )
    {
      getenv_char_p = empty_string;
    }
  assert( strlen(getenv_char_p) < PROCMON_MAX_LEN );
  sprintf( psettings->command, "%s", getenv_char_p );
  /* ----------- */
  /* PROCMON_UID */
  /* ----------- */
  getenv_char_p = getenv( "PROCMON_UID" );
  if( getenv_char_p != NULL )
    {
      psettings->uid = atoi( getenv_char_p );
      if( psettings->interval < 0 )
        {
          psettings->interval = 0;
        }
    }
  /* -------------------------- */
  /* DONE: get env var settings */
  /* -------------------------- */
  /* ------------------------- */
  /* get command line settings */
  /* ------------------------- */
  i = 1;
  while( i < argc && argv != NULL )
    {
      /* ----- */
      /* -time */
      /* ----- */
      if( strcmp( argv[i], "-time" ) == 0 )
        {
          assert( i < argc - 1 );
          psettings->time = atoi( argv[i+1] );
          if( psettings->time < 0 )
            {
              psettings->time = 0;
            }
          i = i + 2;
        }
      /* --------- */
      /* -interval */
      /* --------- */
      else if( strcmp( argv[i], "-interval" ) == 0 )
        {
          assert( i < argc - 1 );
          psettings->interval = atoi( argv[i+1] );
          if( psettings->interval < 0 )
            {
              psettings->interval = 0;
            }
          i = i + 2;
        }
      /* ---- */
      /* -min */
      /* ---- */
      else if( strcmp( argv[i], "-min" ) == 0 )
        {
          assert( i < argc - 1 );
          psettings->min = atoi( argv[i+1] );
          if( psettings->min < 0 )
            {
              psettings->min = 0;
            }
          i = i + 2;
        }
      /* ---- */
      /* -dir */
      /* ---- */
      else if( strcmp( argv[i], "-dir" ) == 0 )
        {
          assert( i < argc - 1 );
          assert( strlen( argv[i+1] ) < PROCMON_MAX_LEN );
          sprintf( psettings->dir, "%s", argv[i+1] );
          i = i + 2;
        }
      /* -------- */
      /* -command */
      /* -------- */
      else if( strcmp( argv[i], "-command" ) == 0 )
        {
          assert( i < argc - 1 );
          assert( strlen( argv[i+1] ) < PROCMON_MAX_LEN );
          sprintf( psettings->command, "%s", argv[i+1] );
          i = i + 2;
        }
      /* ---- */
      /* -uid */
      /* ---- */
      else if( strcmp( argv[i], "-uid" ) == 0 )
        {
          assert( i < argc - 1 );
          psettings->uid = atoi( argv[i+1] );
          if( psettings->interval < 0 )
            {
              psettings->interval = 0;
            }
          i = i + 2;
        }
      /* --- */
      /* pid */
      /* --- */
      else
        {
          psettings->pid = atoi( argv[i] );
          if( psettings->pid == 0 )
            {
              psettings->pid = getpid();
            }
          i = i + 1;
        }
    }
  /* ---------------- */
  /* set other things */
  /* ---------------- */
  if( psettings->pid == 0 &&
      psettings->uid < 0 &&
      strlen( psettings->command ) == 0 )
    {
      psettings->pid = getpid();
    }
  /* -------------------- */
  /* get output file name */
  /* -------------------- */
  psettings->outfile_prefix[0] = '\0';
  psettings->outfile_name[0] = '\0';
  if( strlen( psettings->dir ) > 0 )
    {
      if(gethostname(hostname, PROCMON_MAX_LEN-1) != 0)
        abort();
      sprintf( psettings->outfile_prefix,
               "%s/%s%s%s",
               psettings->dir, "procmon_", hostname, "_" );
      sprintf( psettings->outfile_name,
               "%s%llu%s",
               psettings->outfile_prefix, psettings->pid,".txt" );
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   procmon_info_settings_print

   Purpose
   =======
   Print the data from a PROCMON_INFO_SETTINGS_struct to the specified file.

   Arguments
   =========
   psettings  Intent: in
              C type: const PROCMON_INFO_SETTINGS_struct*
              The struct to print out.
             
   outfile    Intent: in
              C type: FILE*
              The file which to print (eg from fopen).
              If NULL, will print to stdout.
             
   ierr       Intent: out
              C type: (na) error return value
              error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) print settings info.

   ======================================================================== */
int procmon_info_settings_print(
                                const PROCMON_INFO_SETTINGS_struct *psettings,
                                FILE *outfile
                                )
{
  int ierr=0; /* error return value */
  FILE *of; /* pointer to actual output file */
  of = outfile;
  if( outfile == NULL )
    {
      of = stdout;
    }
  if( psettings != NULL )
    {
      fprintf( of,
               "#PROCMON_SETTINGS:"
               " TIME [%llu]"
               " MIN [%llu]"
               " INTERVAL [%llu]"
               " DIR [%s]"
               "\n",
               psettings->time,
               psettings->min,
               psettings->interval,
               psettings->dir );
      fflush( of );
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   procmon_info_copy

   Purpose
   =======
   Copy data from one PROCMON_INFO_struct to another

   Arguments
   =========
   pinfo         Intent: in
                 C type: const PROCMON_INFO_struct*
                 The original pinfo.

   pinfoc        Intent: out
                 C type: PROCMON_INFO_struct*
                 The copy.

   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) copy data

   ======================================================================== */
int procmon_info_copy(
                      const PROCMON_INFO_struct *pinfo,
                      PROCMON_INFO_struct *pinfoc
                      )
{
  int ierr=0; /* error return value */
  int i; /* loop var */
  pinfoc->num_samples    = pinfo->num_samples;
  pinfoc->printed_header = pinfo->printed_header;
  pinfoc->wall           = pinfo->wall;
  pinfoc->m_size         = pinfo->m_size;
  pinfoc->m_free         = pinfo->m_free;
  pinfoc->p_pid          = pinfo->p_pid;
  pinfoc->p_ppid         = pinfo->p_ppid;
  pinfoc->p_uid          = pinfo->p_uid;
  strncpy( pinfoc->p_command, pinfo->p_command, PROCMON_MAX_LEN );
  pinfoc->thread         = pinfo->thread;
  pinfoc->p_size_lim     = pinfo->p_size_lim;
  pinfoc->p_rss_lim      = pinfo->p_rss_lim;
  pinfoc->p_data_lim     = pinfo->p_data_lim;
  pinfoc->p_stack_lim    = pinfo->p_stack_lim;
  pinfoc->p_rss          = pinfo->p_rss;
  pinfoc->p_data         = pinfo->p_data;
  pinfoc->p_stack        = pinfo->p_stack;
  for( i = 0; i < PROCMON_SAMPLE_SAVE; i++ )
    {
      pinfoc->p_size[i] = pinfo->p_size[i];
      pinfoc->p_time[i] = pinfo->p_time[i];
      pinfoc->p_wall[i] = pinfo->p_wall[i];
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   procmon_info_init

   Purpose
   =======
   Initialize a PROCMON_INFO_struct

   Arguments
   =========
   pinfo         Intent: out
                 C type: PROCMON_INFO_struct*
                 The original pinfo.

   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) copy data

   ======================================================================== */
int procmon_info_init(
                      PROCMON_INFO_struct *pinfo
                      )
{
  int ierr = 0; /* error return value */
  int i; /* loop var */
  pinfo->num_samples = 0;
  pinfo->printed_header = 0;
  pinfo->wall = 0;
  pinfo->m_size = 0;
  pinfo->m_free = 0;
  pinfo->p_pid = 0;
  pinfo->p_ppid = 0;
  pinfo->p_uid = 0;
  memset( pinfo->p_command, '\0', PROCMON_MAX_LEN );
  pinfo->thread = 0;
  pinfo->p_size_lim = 0;
  pinfo->p_rss_lim = 0;
  pinfo->p_data_lim = 0;
  pinfo->p_stack_lim = 0;
  pinfo->p_rss = 0;
  pinfo->p_data = 0;
  pinfo->p_stack = 0;
  for( i = 0; i < PROCMON_SAMPLE_SAVE; i++ )
    {
      pinfo->p_size[i] = 0;
      pinfo->p_time[i] = 0;
      pinfo->p_wall[i] = 0;
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   procmon_info_print

   Purpose
   =======
   Print the data from a PROCMON_INFO_struct to the specified file.

   Arguments
   =========
   pinfo         Intent: inout
                 C type: const PROCMON_INFO_struct*
                 The process info struct (filled by procmon_info_get).

   outfile       Intent: in
                 C type: FILE*
                 The file which to print (eg from fopen).
                 If NULL, will print to stdout.

   ierr          Intent: out
                 C type: (na) error return value
                 error return value

   Return Values
   =============
   0 if ok

   Program Flow
   ============
   1) print header or print pinfo.

   ======================================================================== */
int procmon_info_print(
                       PROCMON_INFO_struct *pinfo,
                       FILE *outfile
                       )
{
  int i; /* loop var */
  int ierr=0; /* error return value */
  FILE *of; /* pointer to actual output file */
  double pTimeCur[PROCMON_SAMPLE_SAVE];
  /* ---- */
  /* init */
  /* ---- */
  of = outfile;
  if( outfile == NULL )
    {
      of = stdout;
    }
  /* ------------ */
  /* print header */
  /* ------------ */
  if( pinfo->printed_header == 0 )
    {
      /* ----------- */
      /* static info */
      /* ----------- */
      fprintf( of,
               "#PROCMON:"
               " PID [%llu]"
               " UID [%llu]"
               " MMEMSIZE [%9.2e]"
               " COMMAND [%s]"
               "\n",
               pinfo->p_pid,
               pinfo->p_uid,
               pinfo->m_size,
               pinfo->p_command );
      /* --- */
      /* key */
      /* --- */
      fprintf( of,
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n"
               "#PROCMON:key %9s = %s\n",
               "EpochTime", "Time (seconds) Since Epoch",
               "pWallTime", "Process Wall Time (seconds)",
               "%pTimeTot", "% Wall Time of User + System",
               "%pTimeCur", "% Current Sample Time of User + System",
               "%mMemFree", "% Machine Memory Free",
               "pSize",     "Process Size (bytes)",
               "pSizeMax",  "Maximum Process Size Recorded",
               "%RSS:pSze", "% of Process Size   of RSS",
               "%RSS:mMem", "% of Machine Memory of RSS",
               "%pLimSize", "% of pSize Limit",
               "%pLimRSS",  "% of RSS   Limit",
               "%pLimData", "% of Data  Limit",
               "%pLimStck", "% of Stack Limit",
               "pSizeRSS",  "Process RSS   Size (bytes)",
               "pSizeData", "Process Data  Size (bytes)",
               "pSizeStck", "Process Stack Size (bytes)"
               );
      /* ------------- */
      /* column header */
      /* ------------- */
      fprintf( of,
               "%9s "
               "%17s "
               "%9s "
               "%9s "
               "%9s "
               "%9s "
               "%9s "
               "%9s "
               "%9s "
               "%9s "
               "%9s "
               "%9s "
               "%9s "
               "%9s "
               "%s\n",
               "#PROCMON ",
               "EpochTime",
               "pWallTime",
               "%pTimeTot",
               "%pTimeCur",
               "%mMemFree",
               "pSize",
               "pSizeMax",
               "%RSS:pSze",
               "%RSS:mMem",
               pinfo->p_size_lim   > 0 ? "%pLimSize" : "pSize",
               pinfo->p_rss_lim    > 0 ? "%pLimRSS"  : "pSizeRSS",
               pinfo->p_data_lim   > 0 ? "%pLimData" : "pSizeData",
               pinfo->p_stack_lim  > 0 ? "%pLimStck" : "pSizeStck",
               "" );
      fflush( of );
      pinfo->printed_header = 1;
    }
  /* ---------- */
  /* print info */
  /* ---------- */
  if( pinfo->wall > 0 )
    {
      pTimeCur[0] = 100.0*(pinfo->p_time[0]/pinfo->p_wall[0]);
      for( i = 1; i < PROCMON_SAMPLE_SAVE; i++ )
        {
          if( pinfo->p_wall[i-1]-pinfo->p_wall[i] == 0 )
            {
              pTimeCur[i] = 0;
            }
          else
            {
              pTimeCur[i] = 100.0*((pinfo->p_time[i-1]-pinfo->p_time[i])/
                                   (pinfo->p_wall[i-1]-pinfo->p_wall[i]));
            }
        }
      fprintf( of,
               "%9s "
               "%17f "
               "%9.2e "
               "%9.2e "
               "%9.2e "
               "%9.2e "
               "%9.2e "
               "%9.2e "
               "%9.2e "
               "%9.2e "
               "%9.2e "
               "%9.2e "
               "%9.2e "
               "%9.2e "
               "%s\n",
               "PROCMON  ",
               pinfo->wall,
               pinfo->p_wall[0],
               pTimeCur[0],
               pTimeCur[1],
               pinfo->m_size > 0 ?
               100.0*(pinfo->m_free/pinfo->m_size) : 0,
               pinfo->p_size[0],
               pinfo->p_size[1],
               pinfo->p_size[0] > 0 ?
               100.0*(pinfo->p_rss/pinfo->p_size[0]) : 0,
               pinfo->m_size > 0 ?
               100.0*(pinfo->p_rss/pinfo->m_size) : 0,
               pinfo->p_size_lim > 0 ?
               100.0*(pinfo->p_size[0]/pinfo->p_size_lim) : pinfo->p_size[0],
               pinfo->p_rss_lim > 0 ?
               100.0*(pinfo->p_rss/pinfo->p_rss_lim) : pinfo->p_rss,
               pinfo->p_data_lim > 0 ?
               100.0*(pinfo->p_data/pinfo->p_data_lim) : pinfo->p_data,
               pinfo->p_stack_lim > 0 ?
               100.0*(pinfo->p_stack/pinfo->p_stack_lim) : pinfo->p_stack,
               "" );
      fflush( of );
    }
  return( ierr );
}
/* ========================================================================

   Name
   ====
   procmon_info_get

   Purpose
   =======
   Fill the PROCMON_INFO_struct for the specified pid.

   Arguments
   =========
   pid           Intent: in
                 C type: long long
                 The process id.  If 0, use this processes id.

   pinfo         Intent: inout
                 C type: PROCMON_INFO_struct*
                 The process info struct
                 If NULL, just the header will print out.
                 If there is an error (eg the process went away)
                 this struct will not be modified.

   ierr          Intent: out
                 C type: (na) error return value
                 error return value.
                 If there is an error, pinfo will not be modified.
                 An error might happen if the process cannot be found
                 (ie it terminated) or bad data was detected (eg
                 having process time exceed wall clock time).
                 See the Return Values below.

   Notes
   =====
   This is a collection of machine dependent procmon_info_get routines.
   A good deal of this information was obtained by examining the top utility
   from http://sourceforge.net/projects/unixtop/:
     - Concept of creating a machine independent data structure and a
       routine to fill this data.  Machine dependent fill routines are
       called.
     - Information about how to fill the data was mimicked (or coppied) from
       unixtop routines.

   Return Values
   =============
   0 if ok
   PROCMON_ERROR:   Will likely never fill data (eg process gone)
   PROCMON_WARNING: Current sampling is strange (pinfo not filled)

   Program Flow
   ============
   1) fill stuff from a myriad of sources (/proc, ioctl, sysmp, table,
      getrusage, ...)

   ======================================================================== */
/* ---- */
/* IRIX */
/* ---- */
#if defined( _PROCMON_IRIX )
#include <sys/types.h>
#include <sys/sysmp.h>
#include <sys/sysinfo.h>
int procmon_info_get(
                     long long pid,
                     PROCMON_INFO_struct *pinfo
                     )
{
  int ierr=0; /* error return value */
  PROCMON_INFO_struct pinfot; /* if no error, will be coppied to pinfo */
  char proc_buf_filename[PROCMON_MAX_LEN]; /* filename for /proc file */
  int proc_fid; /* ioctl file id */
  unsigned long long pagesize;
  struct rlimit rlim; /* getrlimit */
  struct rminfo	rminfo_sysmp; /* sysmp */
  struct rusage ruse_getrusage; /* getrusage */
  prpsinfo_t
    prpsinfo_info; /* ioctl(fd, PIOCPSINFO, &prpsinfo_info) */
  prusage_t
    prusage_info; /* ioctl(fd, PIOCUSAGE,  &prusage_info) */
  /* init */
  procmon_info_init( &pinfot );
  pagesize = getpagesize();
  pinfot.p_pid = pid;
  if( pid == 0 )
    {
      pinfot.p_pid = getpid();
    }
  getrlimit( RLIMIT_AS, &rlim );
  pinfot.p_size_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_RSS, &rlim );
  pinfot.p_rss_lim   = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_DATA, &rlim );
  pinfot.p_data_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_STACK, &rlim );
  pinfot.p_stack_lim = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  /* sysmp */
  sysmp(MP_SAGET, MPSA_RMINFO, &rminfo_sysmp, sizeof( struct rminfo ));
  pinfot.m_size = rminfo_sysmp.physmem * pagesize;
  pinfot.m_free = rminfo_sysmp.freemem * pagesize;
  /* /proc/<PID> */
  sprintf( proc_buf_filename, "/proc/%llu", pinfot.p_pid );
  proc_fid = open( proc_buf_filename, O_RDONLY );
  if( proc_fid == -1 )
    {
      ierr = PROCMON_ERROR; /* process is gone */
      pinfot.p_ppid     = 0;
      pinfot.p_size[0]  = 0;
      pinfot.p_rss      = 0;
      pinfot.p_time[0]  = 0;
      pinfot.wall       = 0;
      pinfot.p_wall[0]  = 0;
    }
  else
    {
      ioctl( proc_fid, PIOCPSINFO, &prpsinfo_info );
      ioctl( proc_fid, PIOCUSAGE, &prusage_info );
      close( proc_fid );
      pinfot.p_ppid     = prpsinfo_info.pr_ppid;
      pinfot.p_uid      = prpsinfo_info.pr_uid;
      strncpy( pinfot.p_command, prpsinfo_info.pr_fname, PROCMON_MAX_LEN-1 );
      pinfot.p_size[0]  = prusage_info.pu_size * pagesize;
      pinfot.p_rss      = prusage_info.pu_rss * pagesize;
      pinfot.p_time[0]  = PROCMON_TIMEN_TO_DOUBLE( prusage_info.pu_utime ) + 
        PROCMON_TIMEN_TO_DOUBLE( prusage_info.pu_stime );
      pinfot.wall       = PROCMON_TIMEN_TO_DOUBLE( prusage_info.pu_tstamp );
      pinfot.p_wall[0]  = pinfot.wall - 
        PROCMON_TIMEN_TO_DOUBLE( prusage_info.pu_starttime );
    }
  /* if getting info about this process, can call getrusage to get info */
  if( pid == 0 )
    {
      /* I thought this would work...just returns 0 though... */
      getrusage( RUSAGE_SELF, &ruse_getrusage );
      pinfot.p_data  = ruse_getrusage.ru_idrss;
      pinfot.p_stack = ruse_getrusage.ru_isrss;
    }
  else
    {
      pinfot.p_data  = 0;
      pinfot.p_stack = 0;
    }
  /* ----------------------------------------- */
  /* grab previous sampling values if possible */
  /* ----------------------------------------- */
  if( pinfo->num_samples != 0 )
    {
      pinfot.num_samples    = pinfo->num_samples;
      pinfot.printed_header = pinfo->printed_header;
      pinfot.p_time[1] = pinfo->p_time[0];
      pinfot.p_wall[1] = pinfo->p_wall[0];
      pinfot.p_size[1] = pinfo->p_size[1];
    }
  if( pinfot.p_size[0] > pinfot.p_size[1] )
    {
      pinfot.p_size[1] = pinfot.p_size[0];
    }
  /* ----------------------------------------------------------------- */
  /* on my trials, bad data occurred when the process was almost over  */
  /* and sampling was done via standalone procmon.  So, try and detect */
  /* "bad" data.                                                       */
  /* ----------------------------------------------------------------- */
  if( ( ierr != PROCMON_ERROR ) &&
      ( ( pinfot.p_time[0] > pinfot.p_wall[0] ) ||
        ( (pinfot.p_time[0] - pinfot.p_time[1]) >
          (pinfot.p_wall[0] - pinfot.p_wall[1])*1.1 ) ||
        ( pinfot.p_size == 0 ) ) )
    {
      ierr = PROCMON_WARNING;
    }
  if( ierr == 0 )
    {
      pinfot.num_samples++;
      procmon_info_copy( &pinfot, pinfo );
    }
  return( ierr );
}
#endif /* _PROCMON_IRIX */
/* ----- */
/* ALPHA */
/* ----- */
#if defined( _PROCMON_ALPHA )
#include <sys/table.h>
#include <sys/user.h>
int procmon_info_get(
                     long long pid,
                     PROCMON_INFO_struct *pinfo
                     )
{
  int ierr=0; /* error return value */
  PROCMON_INFO_struct pinfot; /* if no error, will be coppied to pinfo */
  int lerr; /* local error return value */
  char proc_buf_filename[PROCMON_MAX_LEN]; /* filename for /proc file */
  int proc_fid; /* ioctl file id */
  unsigned long long pagesize;
  struct rlimit rlim; /* getrlimit */
  struct user user_tbl; /* table - TBL_UAREA */
  struct tbl_pmemstats pmemstats_tbl; /* table - TBL_PMEMSTATS */
  struct tbl_vmstats vmstats_tbl; /* table - TBL_VMSTATS */
  prpsinfo_t
    prpsinfo_info; /* ioctl(fd, PIOCPSINFO, &prpsinfo_info) */
  prusage_t
    prusage_info; /* ioctl(fd, PIOCUSAGE,  &prusage_info) */
  /* init */
  procmon_info_init( &pinfot );
  pagesize = getpagesize();
  pinfot.p_pid = pid;
  if( pid == 0 )
    {
      pinfot.p_pid = getpid();
    }
  getrlimit( RLIMIT_AS, &rlim );
  pinfot.p_size_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_RSS, &rlim );
  pinfot.p_rss_lim   = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_DATA, &rlim );
  pinfot.p_data_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_STACK, &rlim );
  pinfot.p_stack_lim = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  /* table: TBL_PMEMSTATS */
  table( TBL_PMEMSTATS, 0, &pmemstats_tbl, 1,
         sizeof( struct tbl_pmemstats ) );
  pinfot.m_size = pmemstats_tbl.physmem;
  /* table: TBL_VMSTATS */
  table( TBL_VMSTATS, 0, &vmstats_tbl, 1,
         sizeof( struct tbl_vmstats ) );
  pinfot.m_free = pagesize * ( vmstats_tbl.free_count );
  /* table: TBL_UAREA */
  /* ---------------------------------------------------------------- */
  /* when accessing another pid (eg not this process), table seems to */
  /* return a -1.  So, do not use this.                               */
  /* ---------------------------------------------------------------- */
  lerr = table( TBL_UAREA, pinfot.p_pid, &user_tbl, 1, sizeof( struct user ) );
  if( lerr == -1 )
    {
      pinfot.p_data    = 0;
      pinfot.p_stack   = 0;
      pinfot.p_size[0] = 0;
    }
  else
    {
      pinfot.p_data  = user_tbl.u_dsize * pagesize;
      pinfot.p_stack = user_tbl.u_ssize * pagesize;
      /* use ioclt call below for getting this - works for getting info */
      /* about another pid                                              */
      /* 
         pinfot.p_size  = user_tbl.u_tsize * pagesize +
         pinfot.p_data + pinfot.p_stack;
      */
    }
  /* /proc/<PID> */
  sprintf( proc_buf_filename, "/proc/%llu", pinfot.p_pid );
  proc_fid = open( proc_buf_filename, O_RDONLY );
  if( proc_fid == -1 )
    {
      ierr = PROCMON_ERROR; /* process is gone */
      pinfot.p_ppid    = 0;
      pinfot.p_rss     = 0;
      pinfot.p_time[0] = 0;
      pinfot.wall      = 0;
      pinfot.p_wall[0] = 0;
    }
  else
    {
      ioctl( proc_fid, PIOCPSINFO, &prpsinfo_info );
      ioctl( proc_fid, PIOCUSAGE, &prusage_info );
      close( proc_fid );
      pinfot.p_ppid    = prpsinfo_info.pr_ppid;
      pinfot.p_uid     = prpsinfo_info.pr_uid;
      strncpy( pinfot.p_command, prpsinfo_info.pr_fname, PROCMON_MAX_LEN-1 );
      pinfot.p_size[0] = prpsinfo_info.pr_size * pagesize;
      pinfot.p_rss     = prpsinfo_info.pr_rssize * pagesize;
      pinfot.p_time[0] = PROCMON_TIMEN_TO_DOUBLE( prpsinfo_info.pr_time );
      pinfot.wall      = PROCMON_TIMEN_TO_DOUBLE( prusage_info.pr_tstamp );
      pinfot.p_wall[0] = pinfot.wall - 
        PROCMON_TIMEN_TO_DOUBLE( prusage_info.pr_create );
    }
  /* ----------------------------------------- */
  /* grab previous sampling values if possible */
  /* ----------------------------------------- */
  if( pinfo->num_samples != 0 )
    {
      pinfot.num_samples    = pinfo->num_samples;
      pinfot.printed_header = pinfo->printed_header;
      pinfot.p_time[1] = pinfo->p_time[0];
      pinfot.p_wall[1] = pinfo->p_wall[0];
      pinfot.p_size[1] = pinfo->p_size[1];
    }
  if( pinfot.p_size[0] > pinfot.p_size[1] )
    {
      pinfot.p_size[1] = pinfot.p_size[0];
    }
  /* ----------------------------------------------------------------- */
  /* on my trials, bad data occurred when the process was almost over  */
  /* and sampling was done via standalone procmon.  So, try and detect */
  /* "bad" data.                                                       */
  /* ----------------------------------------------------------------- */
  if( ( ierr != PROCMON_ERROR ) &&
      ( ( pinfot.p_time[0] > pinfot.p_wall[0] ) ||
        ( (pinfot.p_time[0] - pinfot.p_time[1]) >
          (pinfot.p_wall[0] - pinfot.p_wall[1])*1.1 ) ||
        ( pinfot.p_size == 0 ) ) )
    {
      ierr = PROCMON_WARNING;
    }
  if( ierr == 0 )
    {
      pinfot.num_samples++;
      procmon_info_copy( &pinfot, pinfo );
    }
  return( ierr );
}
#endif /* _PROCMON_ALPHA */
/* ----- */
/* LINUX */
/* ----- */
#if defined( _PROCMON_LINUX )
#define JIFFIE 100.0
#define KBYTE 1024.0
#include <ctype.h>
static inline char* skip_token(
                               const char *p
                               )
{
  while (p && isspace(*p)) p++;
  while (p && !isspace(*p)) p++;
  while (p && isspace(*p)) p++;
  return (char *)p;
}
static inline char* skip_ws(
                            const char *p
                            )
{
  while (p && isspace(*p)) p++;
  return (char *)p;
}
static inline long long read_buf(
                                 FILE *filep,
                                 char **file_buf,
                                 long long *file_buf_size
                                 )
{
  long long i = 0;
  do
    {
      if( i >= *file_buf_size -1 )
        {
          *file_buf_size += getpagesize();
          *file_buf = realloc( *file_buf, *file_buf_size );
        }
    }
  while( ((*file_buf)[i++] = fgetc(filep)) != EOF );
  if( i > 0 )
    {
      (*file_buf)[i-1] = '\000';
    }
  return( i );
}
int procmon_info_get(
                     long long pid,
                     PROCMON_INFO_struct *pinfo
                     )
{
  int ierr=0; /* error return value */
  PROCMON_INFO_struct pinfot; /* temp coppied if no error */
  char proc_buf_filename[PROCMON_MAX_LEN]; /* filename for /proc file */
  char *proc_buf; /* /proc name and file contents */
  long long num_chars; /* number of chars to print in strncpy */
  long long proc_buf_size; /* size of buffer */
  char *p; /* pointer into proc_buf */
  char *p_tmp; /* temp pointer into proc_buf */
  unsigned long long pagesize;
  struct rlimit rlim; /* getrlimit */
  struct timespec time_current; /* clock_gettime */
  FILE *proc_fp; /* file pointer */
  double uptime;
  /* init */
  proc_buf_size = 0;
  proc_buf = NULL;
  procmon_info_init( &pinfot );
  pagesize = getpagesize();
  pinfot.p_pid = pid;
  if( pid == 0 )
    {
      pinfot.p_pid = getpid();
    }
  clock_gettime( CLOCK_REALTIME, &time_current );
  pinfot.wall = PROCMON_TIMEN_TO_DOUBLE( time_current );
  getrlimit( RLIMIT_AS, &rlim );
  pinfot.p_size_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_RSS, &rlim );
  pinfot.p_rss_lim   = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_DATA, &rlim );
  pinfot.p_data_lim  = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  getrlimit( RLIMIT_STACK, &rlim );
  pinfot.p_stack_lim = rlim.rlim_cur == RLIM_INFINITY ? -1.0 : rlim.rlim_cur;
  /* ------------- */
  /* /proc/meminfo */
  /* ------------- */
  sprintf( proc_buf_filename, "%s/%s", "/proc", "meminfo" );
  proc_fp = fopen( proc_buf_filename, "r" );
  if( proc_fp == NULL )
    {
      ierr = PROCMON_ERROR; /* cannot read info */
      pinfot.m_size = 0;
      pinfot.m_free = 0;
    }
  else
    {
      read_buf( proc_fp, &proc_buf, &proc_buf_size );
      fclose( proc_fp );
      p = proc_buf;
      /* process buffer */
      p = p != NULL ? strchr( p, '\n' ) : NULL;       /* end of line */
      p = skip_token( p );                            /* Mem: */
      p = skip_ws( p );                               /* spaces */
      pinfot.m_size = p != NULL ? strtod( p, &p ) : 0;/* total -> m_size */ 
      p = skip_token( p );                            /* used */
      pinfot.m_free = p != NULL ? strtod( p, &p ) : 0;/* free -> m_free */
    }
  /* ------------------ */
  /* /proc/<pid>/status */
  /* ------------------ */
  sprintf( proc_buf_filename, "%s/%llu/%s", "/proc", pinfot.p_pid, "status" );
  proc_fp = fopen( proc_buf_filename, "r" );
  if( proc_fp == NULL )
    {
      ierr = PROCMON_ERROR; /* process is gone */
      pinfot.p_ppid    = 0;
      pinfot.p_size[0] = 0;
      pinfot.p_rss     = 0;
      pinfot.p_data    = 0;
      pinfot.p_stack   = 0;
    }
  else
    {
      read_buf( proc_fp, &proc_buf, &proc_buf_size );
      fclose( proc_fp );
      p = proc_buf;
      /* process buffer */
      p = p != NULL ? strstr(p, "Name:" ) : NULL;                 /* Name */
      p = skip_token(p);
      p_tmp = p != NULL ? strchr(p, '\n') : NULL;
      num_chars = p_tmp - p;
      num_chars =
        num_chars < PROCMON_MAX_LEN - 1 ? num_chars : PROCMON_MAX_LEN;
      strncpy( pinfot.p_command, p, num_chars );
      p = p != NULL ? strchr(p, '\n') : NULL;         /* get passed Name */
      p = p != NULL ? strstr(p, "PPid:" ) : NULL;                 /* PPid */
      p = skip_token(p);
      pinfot.p_ppid    = p != NULL ? strtoul( p, &p, 10 ) : 0;
      p = p != NULL ? strstr(p, "Uid:" ) : NULL;                  /* Uid */
      p = skip_token(p);
      pinfot.p_uid    = p != NULL ? strtoul( p, &p, 10 ) : 0;
      p = p != NULL ? strstr(p, "VmSize:" ) : NULL;               /* VmSize */
      p = skip_token(p);
      pinfot.p_size[0] = p != NULL ? strtod( p, &p ) * KBYTE : 0;
      p = p != NULL ? strstr(p, "VmRSS:" ) : NULL;                /* VmRSS */
      p = skip_token(p);
      pinfot.p_rss     = p != NULL ? strtod( p, &p ) * KBYTE : 0;
      p = p != NULL ? strstr(p, "VmData:" ) : NULL;               /* VmData */
      p = skip_token(p);
      pinfot.p_data    = p != NULL ? strtod( p, &p ) * KBYTE : 0;
      p = p != NULL ? strstr(p, "VmStk:" ) : NULL;                /* VmStk */
      p = skip_token(p);
      pinfot.p_stack   = p != NULL ? strtod( p, &p ) * KBYTE : 0;
    }
  /* ---------------- */
  /* /proc/uptime     */
  /* needed by status */
  /* ---------------- */
  sprintf( proc_buf_filename, "%s/%s", "/proc", "uptime" );
  proc_fp = fopen( proc_buf_filename, "r" );
  if( proc_fp == NULL )
    {
      ierr = PROCMON_ERROR; /* cannot read info */
      uptime = 0;
    }
  else
    {
      read_buf( proc_fp, &proc_buf, &proc_buf_size );
      fclose( proc_fp );
      p = proc_buf;
      /* process buffer */
      uptime = p != NULL ? strtod( p, &p ) : 0; /* time since reboot */
    }
  /* ------------------------------------------ */
  /* /proc/<pid>/stat                           */
  /* Most not used - status (above) used mostly */
  /* ------------------------------------------ */
  sprintf( proc_buf_filename, "%s/%llu/%s", "/proc", pinfot.p_pid, "stat" );
  proc_fp = fopen( proc_buf_filename, "r" );
  if( proc_fp == NULL )
    {
      ierr = PROCMON_ERROR; /* process is gone */
      pinfot.p_time[0] = 0;
      pinfot.p_wall[0] = 0;
      pinfot.p_size[0] = 0;
      pinfot.p_rss     = 0;
    }
  else
    {
      read_buf( proc_fp, &proc_buf, &proc_buf_size );
      fclose( proc_fp );
      p = proc_buf;
      /* process buffer */
      p = skip_token(p);                               /* 1  pid */
      p = p != NULL ? strchr(p, ')') + 1 : NULL;       /* 2  command */
      p = skip_token(p);                               /* 3  state */
      p = skip_token(p);                               /* 4  ppid */
      p = skip_token(p);                               /* 5  pgrp */
      p = skip_token(p);                               /* 6  session */
      p = skip_token(p);                               /* 7  tty */
      p = skip_token(p);                               /* 8  tpgid */
      p = skip_token(p);                               /* 9  flags */
      p = skip_token(p);                               /* 10 minflt */
      p = skip_token(p);                               /* 11 cminflt */
      p = skip_token(p);                               /* 12 majflt */
      p = skip_token(p);                               /* 13 cmajflt */
      pinfot.p_time[0]  =
        p != NULL ? strtod( p, &p )/JIFFIE : 0;        /* 14 utime */
      pinfot.p_time[0] +=
        p != NULL ? strtod( p, &p )/JIFFIE : 0;        /* 15 stime */
      p = skip_token(p);                               /* 16 cutime */
      p = skip_token(p);                               /* 17 cstime */
      p = skip_token(p);                               /* 18 counter */
      p = skip_token(p);                               /* 19 priority */
      p = skip_token(p);                               /* 20 timeout */
      p = skip_token(p);                               /* 21 itrealvalue */
      pinfot.p_wall[0]  =
        p != NULL ? -strtod( p, &p )/JIFFIE : 0;       /* 22 starttime */
      pinfot.p_wall[0] += uptime;
      pinfot.p_size[0]  =
        p != NULL ? strtod( p, &p ) : 0;               /* 23 vsize -> p_size */
      pinfot.p_rss   =
        p != NULL ? strtod( p, &p ) * pagesize : 0;    /* 24 rss -> p_rss */
    }
  /* ----------------------------------------- */
  /* grab previous sampling values if possible */
  /* ----------------------------------------- */
  if( pinfo->num_samples != 0 )
    {
      pinfot.num_samples    = pinfo->num_samples;
      pinfot.printed_header = pinfo->printed_header;
      pinfot.p_time[1] = pinfo->p_time[0];
      pinfot.p_wall[1] = pinfo->p_wall[0];
      pinfot.p_size[1] = pinfo->p_size[1];
    }
  if( pinfot.p_size[0] > pinfot.p_size[1] )
    {
      pinfot.p_size[1] = pinfot.p_size[0];
    }
  /* ----------------------------------------------------------------- */
  /* on my trials, bad data occurred when the process was almost over  */
  /* and sampling was done via standalone procmon.  So, try and detect */
  /* "bad" data.                                                       */
  /* ----------------------------------------------------------------- */
  if( ( ierr != PROCMON_ERROR ) &&
      ( ( pinfot.p_time[0] > pinfot.p_wall[0] ) ||
        ( (pinfot.p_time[0] - pinfot.p_time[1]) >
          (pinfot.p_wall[0] - pinfot.p_wall[1])*1.1 ) ||
        ( pinfot.p_size == 0 ) ) )
    {
      ierr = PROCMON_WARNING;
    }
  if( ierr == 0 )
    {
      pinfot.num_samples++;
      procmon_info_copy( &pinfot, pinfo );
    }
  return( ierr );
}
#endif /* _PROCMON_LINUX */
