#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "procmon_info.h"
#define SECONDS 3
int main( int argc, char **argv )
{
  int i; /* loop variable */
  int ierr=0; /* error return status */
  PROCMON_INFO_struct pinfo;
  PROCMON_INFO_SETTINGS_struct psettings;
  FILE *output_fd; /* output file */
  int test_manual; /* if doing the manual test */
  int procmon_error; /* error return value for procmon */
  struct timespec
    timespec_info; /* from clock_gettime */
  double
    time_absolute, /* absolute time */
    time_now, /* current time in secs */
    time_start; /* start time in secs for the interval */
  char *buf;
  if( getenv( "PROCMON_TEST_MANUAL" ) )
    {
      test_manual = 1;
    }
  else
    {
      test_manual = 0;
    }
  output_fd = NULL;
  if( test_manual != 0 )
    {
      procmon_info_init( &pinfo );
      procmon_info_settings_get( 0, NULL, &psettings );
      procmon_error = procmon_info_get( psettings.pid, &pinfo );
      if( strlen( psettings.outfile_name ) > 0 )
        {
          output_fd = fopen( psettings.outfile_name, "a" );
        }
      else
        {
          output_fd = NULL;
        }
      procmon_info_settings_print( &psettings, output_fd );
      if( procmon_error == 0 )
        {
          procmon_info_print( &pinfo, output_fd );
        }
    }
  clock_gettime( CLOCK_REALTIME, &timespec_info );
  time_absolute = timespec_info.tv_sec + timespec_info.tv_nsec/1000000000.0;
  time_start = time_absolute;
  time_now = time_absolute;
  i = 0;
  while( time_now - time_absolute < SECONDS )
    {
      buf = malloc( getpagesize() );
      if( buf == NULL )
        {
          printf( "malloc failed\n" );
          ierr = 1;
          break;
        }
      memset( buf, 1, getpagesize() );
      clock_gettime( CLOCK_REALTIME, &timespec_info );
      time_now = timespec_info.tv_sec + timespec_info.tv_nsec/1000000000.0;
      if( test_manual != 0 )
        {
          if( time_now - time_start > (1.0*psettings.time)/1000.0 )
            {
              time_start = time_now;
              procmon_error = procmon_info_get( psettings.pid, &pinfo );
              if( ( procmon_error == 0 &&
                    psettings.interval > 0 &&
                    pinfo.num_samples % psettings.interval == 0 ) )
                {
                  procmon_info_print( &pinfo, output_fd );
                }
            }
        }
      i++;
    }
  printf( "   Allocation Loops: %d\n", i );
  if( test_manual != 0 )
    {
      procmon_error = procmon_info_get( psettings.pid, &pinfo );
      if( procmon_error == 0 )
        {
          procmon_info_print( &pinfo, output_fd );
        }
    }
  if( output_fd != NULL )
    {
      fclose( output_fd );
    }
  return ierr;
}

